
#include "inc.h"

#include "lib.h"
#include "printf.h"
#include "iwram_arm.h"
#include "div_arm.h"
#include "main.h"

#include "casl.h"

//---------------------------------------------------------------------------
//resource.h
extern ST_MODE3_FONT Mode3Font;
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ROM_DATA ST_CASL_LABEL_KEYWORD CaslLabelKeyWord[CASL_MAX_LABEL_KEYWORD_CNT] = {
	{ "LBL1"  },
	{ "LBL2"  },
	{ "LBL3"  },
	{ "LOOP1" },
	{ "LOOP2" },
	{ "OP"    },
	{ "ED"    } ,
};

ROM_DATA ST_CASL_OP_KEYWORD CaslOpKeyWord[CASL_MAX_OP_KEYWORD_CNT] = {
	//@Bꖽ
	{ "NOP"  , OP_TYPE1 , 0x00 },

	{ "LD"   , OP_TYPE2 , 0x10 },
	{ "ST"   , OP_TYPE3 , 0x11 },
	{ "LAD"  , OP_TYPE3 , 0x12 },

	{ "ADDA" , OP_TYPE2 , 0x20 },
	{ "SUBA" , OP_TYPE2 , 0x21 },
	{ "ADDL" , OP_TYPE2 , 0x22 },
	{ "SUBL" , OP_TYPE2 , 0x23 },

	{ "AND"  , OP_TYPE2 , 0x30 },
	{ "OR"   , OP_TYPE2 , 0x31 },
	{ "XOR"  , OP_TYPE2 , 0x32 },

	{ "CPA"  , OP_TYPE2 , 0x40 },
	{ "CPL"  , OP_TYPE2 , 0x41 },

	{ "SLA"  , OP_TYPE3 , 0x50 },
	{ "SRA"  , OP_TYPE3 , 0x51 },

	{ "JMI"  , OP_TYPE4 , 0x61 },
	{ "JNZ"  , OP_TYPE4 , 0x62 },
	{ "JZE"  , OP_TYPE4 , 0x63 },
	{ "JUMP" , OP_TYPE4 , 0x64 },
	{ "JPL"  , OP_TYPE4 , 0x65 },
	{ "JOV"  , OP_TYPE4 , 0x66 },

	{ "PUSH" , OP_TYPE4 , 0x70 },
	{ "POP"  , OP_TYPE5 , 0x71 },

	{ "CALL" , OP_TYPE4 , 0x80 },
	{ "RET"  , OP_TYPE1 , 0x81 },

	{ "SVC"  , OP_TYPE4 , 0xf0 },
	//AZu
	{ "START", OP_TYPE6 , 0x00 },
	{ "END"  , OP_TYPE7 , 0x00 },
	{ "DS"   , OP_TYPE8 , 0x00 },
	{ "DC"   , OP_TYPE9 , 0x00 },
	//}N
	{ "IN"   , OP_TYPE10, 0x00 },
	{ "OUT"  , OP_TYPE11, 0x00 },
	{ "RPUSH", OP_TYPE12, 0x00 },
	{ "RPOP ", OP_TYPE13, 0x00 },
};

ROM_DATA ST_CASL_R_KEYWORD CaslrKeyWord[CASL_MAX_R_KEYWORD_CNT] = {
	{ "GR0" },
	{ "GR1" },
	{ "GR2" },
	{ "GR3" },
	{ "GR4" },
	{ "GR5" },
	{ "GR6" },
	{ "GR7" },
};

ROM_DATA ST_CASL_ADR_KEYWORD CasladrKeyWord[CASL_MAX_ADR_KEYWORD_CNT] = {
	{ "LBL1"  },
	{ "LBL2"  },
	{ "LBL3"  },
	{ "LOOP1" },
	{ "LOOP2" },
	{ "0"     },
	{ "1"     },
	{ "2"     },
	{ "-1"    },
	{ "OP"    },
	{ "ED"    } ,
};

ROM_DATA ST_CASL_X_KEYWORD CaslxKeyWord[CASL_MAX_X_KEYWORD_CNT] = {
//	{ "GR0" },
	{ "GR1" },
	{ "GR2" },
	{ "GR3" },
	{ "GR4" },
	{ "GR5" },
	{ "GR6" },
	{ "GR7" },
};

ROM_DATA ST_CASL_X_KEYWORD CaslnumKeyWord[CASL_MAX_NUM_KEYWORD_CNT] = {
	{ "0" },
	{ "1" },
	{ "2" },
	{ "3" },
};

//---------------------------------------------------------------------------
ST_CASL   Casl   EWRAM_DATA;
ST_CURSOR Cursor EWRAM_DATA;

//---------------------------------------------------------------------------
EWRAM_CODE void CaslInit()
{
	ST_CASL* p = &Casl;

	p->startLine = 0;

	u16 i;
	for(i=0; i<CASL_MAX_LINE_CNT; i++)
	{
		CaslInitLine(i);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslInitLine(u8 wy)
{
	ST_CASL* p = &Casl;
	u16 i;

	for(i=0; i<CASL_LINE_LENGTH - 1; i++)
	{
		p->line[wy].str[i]    = ' ';
	}
	p->line[wy].str[CASL_LINE_LENGTH - 1] = '\0';

	p->line[wy].opType   = OP_TYPE1;
	p->line[wy].lblCnt   = 0 - 1;		//܂낵ȂiOOG
	p->line[wy].opCnt    = 0 - 1;
	p->line[wy].param[0] = 0 - 1;
	p->line[wy].param[1] = 0 - 1;
	p->line[wy].param[2] = 0 - 1;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEdit()
{
	ST_CURSOR* p = &Cursor;

	if(p->key == KEY_A)				//wordInsert
	{
		CaslEditWordInsert();
		CaslDrawLine(p->wy);
	}
	else if(p->key == KEY_B)		//wordDelete
	{
		CaslEditWordDelete();
		CaslDrawLine(p->wy);
	}
	else if(p->key == KEY_L)		//LineDelete
	{
		CaslEditLineDelete();
		CaslDrawLineAll();
	}
	else if(p->key == KEY_R)		//LineInsert
	{
		CaslEditLineInsert();
		CaslDrawLineAll();
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditLineInsert()
{
	ST_CASL*   c  = &Casl;
	ST_CURSOR* p  = &Cursor;
	u16        wy = p->wy;
	u16 i;

	for(i=CASL_MAX_LINE_CNT - 1; i>wy; i--)
	{
		_Memcpy((u8*)&c->line[i], (u8*)&c->line[i-1], sizeof(ST_CASL_LINE));
	}

	//wsɑ}
	CaslInitLine(wy);
	p->selPoint = 2;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditLineDelete()
{
	ST_CASL*   c  = &Casl;
	ST_CURSOR* p  = &Cursor;
	u16        wy = p->wy;
	u16 i;

	//ws̍폜
	CaslInitLine(wy);
	p->selPoint = 2;

	for(i=wy; i<CASL_MAX_LINE_CNT - 1; i++)
	{
		_Memcpy((u8*)&c->line[i], (u8*)&c->line[i+1], sizeof(ST_CASL_LINE));
	}

	CaslInitLine(CASL_MAX_LINE_CNT - 1);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsert()
{
	ST_CURSOR* p = &Cursor;

	//݂̃J[\ʒuҏWꏊ߂܂
	switch(p->selPoint)
	{
	//label
	case 1:
		CaslEditWordInsertLabel();
		break;
	//op
	case 2:
		CaslEditWordInsertOp();
		break;
	//param1 
	case 3:
		CaslEditWordInsertParam1();
		break;
	//param2
	case 4:
		CaslEditWordInsertParam2();
		break;
	//param3
	case 5:
		CaslEditWordInsertParam3();
		break;
	default:
		DebugMessage("Err: CaslEditWordInsert");
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertLabel()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy     = p->wy;
	u8 lblCnt = c->line[wy].lblCnt + 1;

	CaslEditWordDelete();

	if(lblCnt >= CASL_MAX_LABEL_KEYWORD_CNT)
	{
		lblCnt = 0;
	}

	//xRs[
	u8* src = &c->line[wy].str[0];
	u8* dst = CaslLabelKeyWord[lblCnt].word;
	CaslStrcpy(src, dst);

	c->line[wy].lblCnt = lblCnt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertOp()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy     = p->wy;
	u8 opCnt = c->line[wy].opCnt + 1;

	if(opCnt >= CASL_MAX_OP_KEYWORD_CNT)
	{
		opCnt = 0;
	}

	u8* src = &c->line[wy].str[6];
	u8* dst = CaslOpKeyWord[opCnt].word;

	//IyhׁAO̖߃R[hiopCntj͐ɎĂ܂
	CaslEditWordDelete();
	CaslStrcpy(src, dst);

	c->line[wy].opCnt = opCnt;
	c->line[wy].opType = CaslOpKeyWord[opCnt].opType;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertParam1()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy     = p->wy;
	u8 opType = c->line[wy].opType;

	switch(opType)
	{
	//r
	case OP_TYPE2:
	case OP_TYPE3:
	case OP_TYPE5:
		CaslEditWordInsertR();
		break;
	//adr
	case OP_TYPE4:
		CaslEditWordInsertAdr();
		break;
	//DS,DC
	case OP_TYPE8:
	case OP_TYPE9:
		CaslEditWordInsertNum();
		break;
	//IN,OUT
	case OP_TYPE10:
	case OP_TYPE11:
		//TODO
		break;

	default:
		DebugMessage("Err: CaslEditWordInsertParam1");
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertParam2()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy     = p->wy;
	u8 opType = c->line[wy].opType;

	switch(opType)
	{
	//adr
	case OP_TYPE2:
	case OP_TYPE3:
		CaslEditWordInsertRorAdr();
		break;
	//x
	case OP_TYPE4:
		CaslEditWordInsertX();
		break;
	//IN,OUT
	case OP_TYPE10:
	case OP_TYPE11:
		//TODO
		break;

	default:
		DebugMessage("Err: CaslEditWordInsertParam2");
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertParam3()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy     = p->wy;
	u8 opType = c->line[wy].opType;

	switch(opType)
	{
	//x
	case OP_TYPE2:
	case OP_TYPE3:
		CaslEditWordInsertX();
		break;

	default:
		DebugMessage("Err: CaslEditWordInsertParam3");
		break;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertR()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy       = p->wy;
	u8 idx      = p->selPoint - 3;
	u8 paramCnt = c->line[wy].param[idx] + 1;

	if(paramCnt >= CASL_MAX_R_KEYWORD_CNT)
	{
		paramCnt = 0;
	}

	u8 buf[32];
	u8* src = &c->line[wy].str[(p->selPoint-1) * CURSOR_MOVE_X_CNT];
	u8* dst = CaslGetRevisionStrComma(buf, CaslrKeyWord[paramCnt].word);

	CaslEditWordDelete();
	CaslStrcpy(src, dst);

	c->line[wy].param[idx] = paramCnt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertAdr()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy       = p->wy;
	u8 idx      = p->selPoint - 3;
	u8 paramCnt = c->line[wy].param[idx] + 1;

	if(paramCnt >= CASL_MAX_ADR_KEYWORD_CNT)
	{
		paramCnt = 0;
	}

	u8* src = &c->line[wy].str[(p->selPoint-1) * CURSOR_MOVE_X_CNT];
	u8* dst = CasladrKeyWord[paramCnt].word;

	CaslEditWordDelete();
	CaslStrcpy(src, dst);

	c->line[wy].param[idx] = paramCnt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertRorAdr()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8  wy       = p->wy;
	u8  idx      = p->selPoint - 3;
	u8  opType   = c->line[wy].opType;
	u8  paramCnt = c->line[wy].param[idx] + 1;
	u8* src      = &c->line[wy].str[(p->selPoint-1) * CURSOR_MOVE_X_CNT];
	u8* dst;
	u8  buf[32];

	//ur, adr, xvur1, r2v΍
	//adrre[u̓e\܂
	if(opType == OP_TYPE2)
	{
		if(paramCnt >= CASL_MAX_R_KEYWORD_CNT + CASL_MAX_ADR_KEYWORD_CNT)
		{
			paramCnt = 0;
			dst      = CaslrKeyWord[paramCnt].word;
		}
		else if(paramCnt >= CASL_MAX_R_KEYWORD_CNT)
		{
			dst = CasladrKeyWord[paramCnt - CASL_MAX_R_KEYWORD_CNT].word;
		}
		else
		{
			dst = CaslrKeyWord[paramCnt].word;
		}
		dst = CaslGetRevisionStrComma(buf, dst);
	}
	else
	{
		if(paramCnt >= CASL_MAX_ADR_KEYWORD_CNT)
		{
			paramCnt = 0;
			dst      = CasladrKeyWord[paramCnt].word;
		}
		else
		{
			dst = CasladrKeyWord[paramCnt].word;
		}
	}

	CaslEditWordDelete();
	CaslStrcpy(src, dst);

	c->line[wy].param[idx] = paramCnt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertX()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy       = p->wy;
	u8 idx      = p->selPoint - 3;
	u8 paramCnt = c->line[wy].param[idx] + 1;

	if(paramCnt >= CASL_MAX_X_KEYWORD_CNT)
	{
		paramCnt = 0;
	}

	u8 buf[32];
	u8* src = &c->line[wy].str[(p->selPoint-1) * CURSOR_MOVE_X_CNT];
	u8* dst = CaslGetRevisionStrComma(buf, CaslrKeyWord[paramCnt].word);

	CaslEditWordDelete();
	CaslStrcpy(src, dst);

	c->line[wy].param[idx] = paramCnt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordInsertNum()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8 wy       = p->wy;
	u8 idx      = p->selPoint - 3;
	u8 paramCnt = c->line[wy].param[idx] + 1;

	if(paramCnt >= CASL_MAX_NUM_KEYWORD_CNT)
	{
		paramCnt = 0;
	}

	u8* src = &c->line[wy].str[(p->selPoint-1) * CURSOR_MOVE_X_CNT];
	u8* dst = CaslnumKeyWord[paramCnt].word;

	CaslEditWordDelete();
	CaslStrcpy(src, dst);

	c->line[wy].param[idx] = paramCnt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslEditWordDelete()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	u8  wy     = p->wy;
	u8  x      = p->x - p->minX;
	u8* str    = &c->line[wy].str[x];
	u8  delCnt = CASL_WORD_LENGTH;
	u16 i;

	switch(p->selPoint)
	{
	case 1:
		c->line[wy].lblCnt = 0 - 1;
		break;
	case 2:
		//ߌꍇAIyh܂
		delCnt = _Length((u8*)str);
		c->line[wy].opType   = OP_TYPE1;
		c->line[wy].param[0] = 0 - 1;
		c->line[wy].param[1] = 0 - 1;
		c->line[wy].param[2] = 0 - 1;
		c->line[wy].opCnt    = 0 - 1;
		break;
	case 3:
		c->line[wy].param[0] = 0 - 1;
		break;
	case 4:
		c->line[wy].param[1] = 0 - 1;
		break;
	case 5:
		c->line[wy].param[2] = 0 - 1;
		break;
	default:
		DebugMessage("Err: CaslEditWordDelete");
		break;
	}

	for(i=0; i<delCnt; i++)
	{
		*str++ = ' ';
	}
}
//---------------------------------------------------------------------------
//J[\̈ړ(xSelPoint)́Aړ\`FbN܂
EWRAM_CODE bool CaslIsSelectPoint(u8 wy, u8 xSelPoint)
{
	ST_CASL* p = &Casl;

	//uAZu߂ENDv̓IyhȊOAړs
	if(p->line[wy].opType == OP_TYPE7)
	{
		return FALSE;
	}

	//̃`FbN
	if(xSelPoint == 0)
	{
		return FALSE;
	}

	return xSelPoint <= CaslGetMaxSelectPoint(wy);
}
//---------------------------------------------------------------------------
//Lq`̍ő发߂܂
EWRAM_CODE u8 CaslGetMaxSelectPoint(u8 wy)
{
	ST_CASL* p = &Casl;

	switch(p->line[wy].opType)
	{
		case OP_TYPE1:  return 2;
		case OP_TYPE2:  return 5;
		case OP_TYPE3:  return 5;
		case OP_TYPE4:  return 4;
		case OP_TYPE5:  return 3;
		case OP_TYPE6:  return 2;	//START  GɂȂ̂ŕύXB{3
		case OP_TYPE7:  return 2;
		case OP_TYPE8:  return 3;
		case OP_TYPE9:  return 3;
		case OP_TYPE10: return 4;
		case OP_TYPE11: return 4;
		case OP_TYPE12: return 2;
		case OP_TYPE13: return 2;
	}

	DebugMessage("Err: CaslGetMaxDescription");
	return 0;
}
//---------------------------------------------------------------------------
//J[\̈ړʒu̕␳l߂܂
EWRAM_CODE u8 CaslGetRevisionSelectPoint(u8 wy)
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	//END̓IyhȊOAړs
	if(c->line[wy].opType == OP_TYPE7)
	{
		return 2;
	}

	return MIN(p->selPoint, CaslGetMaxSelectPoint(wy));
}
//---------------------------------------------------------------------------
EWRAM_CODE u8* CaslGetRevisionStrComma(u8* dst, u8* src)
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	//݂̃J[\ʒuҏWꏊ߂܂
	switch(p->selPoint)
	{
	case 3:
		_Strcpy(dst, src);
		_Strcat(dst, (u8*)",");
		break;

	case 5:
except:
		_Strcpy(dst, (u8*)",");
		_Strcat(dst, src);
		break;

	case 4:
		if(c->line[p->wy].opType == OP_TYPE4)
		{
			//TODO u,xvp̗OBȏ㉘Ȃ\[XiOOG
			goto except;
		}
		_Strcpy(dst, src);
		break;

	default:
		DebugMessage("Err: CaslGetRevisionStrComma");
		break;
	}

	return dst;
}
//---------------------------------------------------------------------------
//ʂ̃XN[ C̍ĕ`sꍇTRUE
EWRAM_CODE bool CaslScroll()
{
	ST_CURSOR* p = &Cursor;
	ST_CASL*   c = &Casl;

	if(p->key & KEY_DOWN)
	{
		if((c->startLine + CASL_MAX_DRAW_LINE_CNT) >= CASL_MAX_LINE_CNT)
		{
			return FALSE;
		}
		p->wy++;
		c->startLine++;
	}

	if(p->key & KEY_UP)
	{
		if(c->startLine == 0)
		{
			return FALSE;
		}
		p->wy--;
		c->startLine--;
	}

	p->selPoint = CaslGetRevisionSelectPoint(p->wy);
	p->x        = p->minX + (p->selPoint - 1)  * CURSOR_MOVE_X_CNT;

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslDrawLine(u8 wy)
{
	ST_CASL* p = &Casl;
	char buf[64];

	u8 i = wy - p->startLine;

	_Sprintf(buf, "%s", p->line[wy].str);
	Mode3DrawFontBackGround(4, 1 + i, (u8*)buf, COLOR_WHITE, COLOR_BLACK);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslDrawLineAll()
{
	ST_CASL* p = &Casl;
	char buf[64];

	u16 i;
	for(i=0; i<CASL_MAX_DRAW_LINE_CNT; i++)
	{
		_Sprintf(buf, "%s", p->line[p->startLine + i].str);
		Mode3DrawFontBackGround(4, 1 + i, (u8*)buf, COLOR_WHITE, COLOR_BLACK);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslDrawLineNum()
{
	ST_CASL* p = &Casl;
	char buf[64];

	u16 i;
	for(i=0; i<CASL_MAX_DRAW_LINE_CNT; i++)
	{
		_Sprintf(buf, "%3d|", p->startLine + i + 1);
		Mode3DrawFontBackGround(0, 1 + i, (u8*)buf, COLOR_WHITE, COLOR_BLACK);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslLoadSRAM()
{
	u8* rp1       = (u8*)&Casl;
	u8* rp2       = (u8*)&Cursor;
	u32 caslCnt   = sizeof(ST_CASL);
	u32 cursorCnt = sizeof(ST_CURSOR);
	u32 readCnt   = 0;
	u8  readData;
	u32 i;

	static u8 sig[] = {'C','A','S','L','G','B','A'};
	u8 sigCnt = 7;

	//VOl`̃`FbN܂
	for(i=0; i<sigCnt; i++)
	{
		readData = SRAMRead8(readCnt++);

		if(readData != sig[i])
		{
			DebugMessage("Err: CaslLoadSRAM Sig");
		}
	}

	for(i=0; i<caslCnt; i++)
	{
		readData = SRAMRead8(readCnt++);
		*rp1++   = readData;
	}
	for(i=0; i<cursorCnt; i++)
	{
		readData = SRAMRead8(readCnt++);
		*rp2++   = readData;
	}

	//SRAM̈̃`FbN܂
	if(readCnt >= 0x10000)
	{
		DebugMessage("Err: CaslLoadSRAM Cnt");
	}

	TRACEOUT("Load OK [size = %x]\n", readCnt);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CaslSaveSRAM()
{
	u8* wp1       = (u8*)&Casl;
	u8* wp2       = (u8*)&Cursor;
	u32 caslCnt   = sizeof(ST_CASL);
	u32 cursorCnt = sizeof(ST_CURSOR);
	u32 writeCnt  = 0;
	u32 i;

	//VOl`̓o^܂
	SRAMWrite8 (writeCnt++, 'C');
	SRAMWrite8 (writeCnt++, 'A');
	SRAMWrite8 (writeCnt++, 'S');
	SRAMWrite8 (writeCnt++, 'L');
	SRAMWrite8 (writeCnt++, 'G');
	SRAMWrite8 (writeCnt++, 'B');
	SRAMWrite8 (writeCnt++, 'A');

	for(i=0; i<caslCnt; i++)
	{
		SRAMWrite8 (writeCnt++, *wp1++);
	}
	for(i=0; i<cursorCnt; i++)
	{
		SRAMWrite8 (writeCnt++, *wp2++);
	}

	if(writeCnt >= 0x10000)
	{
		DebugMessage("Err: CaslSaveSRAM");
	}

	TRACEOUT("Save OK [size = %x]\n", writeCnt);
}
//---------------------------------------------------------------------------
//CASL2pɕ̏I[L('\0')͓ĂȂ
EWRAM_CODE u8* CaslStrcpy(u8* ret, u8* s2)
{
	u8 *s1 = ret;
	while(*s2 != '\0')
	{
		*s1++ = *s2++;
	}

	return ret;
}












//---------------------------------------------------------------------------
EWRAM_CODE void CursorInit()
{
	ST_CURSOR*     p = &Cursor;
	ST_MODE3_FONT* f = &Mode3Font;

	p->x         = 4;
	p->y         = 1;
	p->wy        = 0;
	p->minX      = 4;
	p->maxX      = 38;
	p->minY      = 1;
	p->maxY      = 12;
	p->cx        = f->cx;
	p->cy        = f->cy;
	p->sprNo     = 0;
	p->sprTile   = CURSOR_FILL_TILENO;
	p->sprSize   = SP_SIZE_16;
	p->sprForm   = SP_SQUARE;
	p->key       = KEY_NONE;
	p->blinkFlag = FALSE;
	p->selPoint  = 1;

	SpriteSetChr (p->sprNo, p->sprTile);
	SpriteSetSize(p->sprNo, p->sprSize, p->sprForm, SP_COLOR_16);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorInput()
{
	ST_CURSOR* p = &Cursor;

	bool isDrawCursor = FALSE;
	bool isDrawLine   = FALSE;

	s16 tx        = p->x;
	s16 ty        = p->y;
	s16 twy       = p->wy;
	u8  tselPoint = p->selPoint;

	if(p->key & KEY_UP   ) { ty--; twy--; }
	if(p->key & KEY_DOWN ) { ty++; twy++; }
	if(p->key & KEY_LEFT ) { tselPoint--; tx-=CURSOR_MOVE_X_CNT; }
	if(p->key & KEY_RIGHT) { tselPoint++; tx+=CURSOR_MOVE_X_CNT; }

	if((p->key & KEY_UP) || (p->key & KEY_DOWN))
	{
		if(ty < p->minY || ty > p->maxY)
		{
			isDrawLine = CaslScroll();
		}
		else
		{
			p->selPoint  = CaslGetRevisionSelectPoint(twy);
			p->x         = p->minX + (p->selPoint - 1)  * CURSOR_MOVE_X_CNT;
			p->y         = ty;
			p->wy        = twy;
			isDrawCursor = TRUE;
		}
	}

	if((p->key & KEY_LEFT) || (p->key & KEY_RIGHT))
	{
		if( CaslIsSelectPoint(p->wy, tselPoint) == TRUE )
		{
			p->x         = tx;
			p->selPoint  = tselPoint;
			isDrawCursor = TRUE;
		}
	}

	if(isDrawLine == TRUE)
	{
		CaslDrawLineAll();
		CaslDrawLineNum();

		CursorDraw();
		CursorDrawPointNum();
	}
	else
	{
		if(isDrawCursor == TRUE)
		{
			CursorDraw();
			CursorDrawPointNum();
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorDraw()
{
	ST_CURSOR* p = &Cursor;
	u16 x = p->x * p->cx;
	u16 y = p->y * p->cy;

	SpriteMove(p->sprNo, x, y);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorDrawBlink()
{
	ST_CURSOR* p = &Cursor;

	//J[\̈ړ́uhԂv\܂
	if(p->key != KEY_NONE)
	{
		p->sprTile   = CURSOR_FILL_TILENO;
		p->blinkFlag = FALSE;

		goto DrawEnd;
	}

	//_ŗp̃EFCg
	if( _UMod(Timer.clockTick ,CURSOR_BLINK_TIME) != 0 )
	{
		return;
	}

	if(p->blinkFlag == TRUE)
	{
		p->sprTile = CURSOR_FILL_TILENO;
	}
	else
	{
		p->sprTile = CURSOR_CLS_TILENO;
	}
	p->blinkFlag ^= TRUE;


DrawEnd:
	SpriteSetChr (p->sprNo, p->sprTile);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CursorDrawPointNum()
{
	ST_CURSOR* p = &Cursor;
	char buf[64];

	_Sprintf(buf, "     [%3d:%3d] ", p->x - p->minX + 1, p->wy + 1);
	Mode3DrawFontBackGround(25, 0, (u8*)buf, COLOR_BLACK, COLOR_WHITE);
}
