
#include "inc.h"

#include "lib.h"
#include "printf.h"
#include "iwram_arm.h"
#include "div_arm.h"
#include "main.h"

#include "casl.h"
#include "comet.h"

//---------------------------------------------------------------------------
//main.c
extern u8 State;
//casl.c
extern ST_CASL Casl;
extern ST_CASL_ADR_KEYWORD CasladrKeyWord[CASL_MAX_ADR_KEYWORD_CNT];
extern ST_CASL_OP_KEYWORD  CaslOpKeyWord[CASL_MAX_OP_KEYWORD_CNT];
extern ST_CASL_X_KEYWORD   CaslxKeyWord[CASL_MAX_X_KEYWORD_CNT];
extern ST_CASL_R_KEYWORD   CaslrKeyWord[CASL_MAX_R_KEYWORD_CNT];

//---------------------------------------------------------------------------
ST_COMET   Comet   EWRAM_DATA;
ST_COMPILE Compile EWRAM_DATA;

//---------------------------------------------------------------------------
EWRAM_CODE void CompileInit()
{
	_Memset((u8*)&Compile, 0x00, sizeof(ST_COMPILE));

	Compile.pMem = (u8*)&Comet.mem;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileInitTokenList()
{
	_Memset((u8*)&Compile.tokenBuf, 0x00, COMPILE_TOKEN_BUF_LENGTH);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileExec()
{
	ST_CASL* c = &Casl;

	u16 i;
	for(i=0; (i < CASL_MAX_LINE_CNT) && (Compile.endFlag != TRUE); i++)
	{
		//g[NXg쐬
		if( CompileMakeTokenList(c->line[i].str) == FALSE )
		{
			TRACEOUT("Err: Token line=[%d]\n", i+1);
			goto Err;
		}
		CompileDebugToken(i);

		//R[h𐶐ROMɔzu
		if( CompileInsertCode() == FALSE )
		{
			TRACEOUT("Err: Code line=[%d]\n", i+1);
			goto Err;
		}
		CompileDebugMemory();
	}

	//x̎QƂ
	if( CompileLinkLabelTable() == FALSE )
	{
		TRACEOUT("Err: Link\n");
		goto Err;
	}
	CompileDebugMemory();

	State = STATE_RUN;
	return;

Err:
	State = STATE_MSG;
	for(;;){}
}
//---------------------------------------------------------------------------
//g[NXg̍쐬A`FbN܂݂܂
EWRAM_CODE bool CompileMakeTokenList(u8* str)
{
	CompileInitTokenList();

	u8* src = str;
	u8* dst = Compile.tokenBuf;

	//Ls̃`FbN
	if( CompileIsStrSpace(src) == TRUE )
	{
		return TRUE;
	}

	//x`FbN
	if( CompileIsChrName(*src) == TRUE )
	{
		CompileStepLabel(&src, &dst);
	}

	//󔒃`FbN
	if( CompileIsChrSpace(*src) == FALSE )
	{
		return FALSE;
	}
	CompileStepSpace(&src);

	//߃R[h`FbN
	if( CompileIsStrOp(src) == FALSE )
	{
		return FALSE;
	}
	u8 opIdx = CompileStepOp(&src, &dst);

	//󔒃`FbN
	if( CompileIsChrSpace(*src) == FALSE )
	{
		return FALSE;
	}
	CompileStepSpace(&src);

	//--------------------------------------------------
	//param1`FbN
	switch( CaslOpKeyWord[opIdx].opType )
	{
	//none
    //START
	//END
	//RPUSH
	//RPOP
	case OP_TYPE1:
	case OP_TYPE6:
	case OP_TYPE7:
	case OP_TYPE12:
	case OP_TYPE13:
		return CompileIsStrSpace(src);

	//r, adr ,xAr1, r2
	//r, adr ,x
	//r
	case OP_TYPE2:
	case OP_TYPE3:
	case OP_TYPE5:
		if( CompileIsStrR(src) == FALSE )
		{
			return FALSE;
		}
		CompileStepR(&src, &dst);
		break;

	//adr ,x
	case OP_TYPE4:
		if( CompileIsStrAdr(src) == FALSE )
		{
			return FALSE;
		}
		CompileStepAdr(&src, &dst);
		break;

	//DS,DC
	case OP_TYPE8:
	case OP_TYPE9:
		if( CompileIsChrNum(*src) == FALSE )
		{
			return FALSE;
		}
		CompileStepNum(&src, &dst);
		return CompileIsStrSpace(src);

	//IN,OUT
	case OP_TYPE10:
	case OP_TYPE11:
		//TODO
		break;

	default:
		DebugMessage("Err: CompileMakeTokenList param1");
		break;
	}
	//--------------------------------------------------
	CompileStepSpace(&src);

	//R}`FbN
	if( CompileIsChrComma(*src) == FALSE )
	{
		return FALSE;
	}
	src++;
	CompileStepSpace(&src);

	//--------------------------------------------------
	//param2`FbN
	switch( CaslOpKeyWord[opIdx].opType )
	{
	//r, adr ,xAr1, r2
	//r, adr ,x
	case OP_TYPE2:
	case OP_TYPE3:
		if( CompileIsStrR(src) == TRUE )
		{
			CompileStepX(&src, &dst);
			return CompileIsStrSpace(src);
		}
		else if( CompileIsStrAdr(src) == TRUE )
		{
			CompileStepAdr(&src, &dst);
		}
		else
		{
			return FALSE;
		}
		break;

	//adr, x
	case OP_TYPE4:
		if( CompileIsStrX(src) == FALSE )
		{
			return FALSE;
		}
		CompileStepX(&src, &dst);
		return CompileIsStrSpace(src);

	//IN,OUT
	case OP_TYPE10:
	case OP_TYPE11:
		//TODO
		break;

	default:
		DebugMessage("Err: CompileMakeTokenList param2");
		break;
	}
	//--------------------------------------------------
	CompileStepSpace(&src);

	//R}`FbN
	if( CompileIsChrComma(*src) == FALSE )
	{
		return CompileIsStrSpace(src);
	}
	src++;
	CompileStepSpace(&src);

	//--------------------------------------------------
	//param3`FbN
	switch( CaslOpKeyWord[opIdx].opType )
	{
	//r, adr ,xAr1, r2
	//r, adr ,x
	case OP_TYPE2:
	case OP_TYPE3:
		if( CompileIsStrX(src) == FALSE )
		{
			return FALSE;
		}
		CompileStepX(&src, &dst);
		break;

	default:
		DebugMessage("Err: CompileMakeTokenList param3");
		break;
	}

	return CompileIsStrSpace(src);
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepSpace(u8** src)
{
	while( CompileIsChrSpace(**src) == TRUE )
	{
		*src += 1;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepNull(u8** src)
{
	while( CompileIsChrNull(**src) != TRUE )
	{
		*src += 1;
	}
	*src += 1;		//'\0'̎w
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepLabel(u8** src, u8** dst)
{
	**dst = COMET_TOKEN_LABEL;
	*dst += 1;

	do
	{
		**dst = **src;
		*src += 1;
		*dst += 1;

	} while( CompileIsChrName(**src) == TRUE || CompileIsChrNum(**src) == TRUE );

	**dst = '\0';
	*dst  += 1;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 CompileStepOp(u8** src, u8** dst)
{
	u8  opWord[COMPILE_WORD_LENGTH];
	u8* p = opWord;

	do
	{
		*p++ = **src;
		*src += 1;

	} while( CompileIsChrName(**src) == TRUE );

	*p++ = '\0';

	u16 i;
	for(i=0; i<CASL_MAX_OP_KEYWORD_CNT; i++)
	{
		if( _Strcmp(opWord, CaslOpKeyWord[i].word) == 0 )
		{
			**dst  = COMET_TOKEN_OP;
			*dst  += 1;
			**dst  = i;
			*dst  += 1;

			return i;
		}
	}

	DebugMessage("Err: CompileStepOp");
	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepR(u8** src, u8** dst)
{
	u8  rWord[COMPILE_WORD_LENGTH];
	u8* p = rWord;

	do
	{
		*p++ = **src;
		*src += 1;

	} while( (CompileIsChrName(**src) == TRUE) || (CompileIsChrNum(**src) == TRUE) );

	*p++ = '\0';

	u16 i;
	for(i=0; i<CASL_MAX_R_KEYWORD_CNT; i++)
	{
		if( _Strcmp(rWord, CaslrKeyWord[i].word) == 0 )
		{
			**dst  = COMET_TOKEN_R;
			*dst  += 1;
			**dst  = i;
			*dst  += 1;

			return;
		}
	}

	DebugMessage("Err: CompileStepR");
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepX(u8** src, u8** dst)
{
	u8  xWord[COMPILE_WORD_LENGTH];
	u8* p = xWord;

	do
	{
		*p++ = **src;
		*src += 1;

	} while( (CompileIsChrName(**src) == TRUE) || (CompileIsChrNum(**src) == TRUE) );

	*p++ = '\0';

	u16 i;
	for(i=0; i<CASL_MAX_X_KEYWORD_CNT; i++)
	{
		if( _Strcmp(xWord, CaslxKeyWord[i].word) == 0 )
		{
			**dst  = COMET_TOKEN_X;
			*dst  += 1;
			**dst  = i;
			*dst  += 1;

			return;
		}
	}

	DebugMessage("Err: CompileStepX");
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepAdr(u8** src, u8** dst)
{
	u8  adrWord[COMPILE_WORD_LENGTH];
	u8* p = adrWord;

	do
	{
		*p++ = **src;
		*src += 1;

	} while( (CompileIsChrName(**src) == TRUE) || (CompileIsChrNum(**src) == TRUE) );

	*p++ = '\0';

	//݂̂ō\Ă邩`FbN܂
	if( CompileIsStrNum(adrWord) == TRUE )
	{
		**dst  = COMET_TOKEN_ADR_NUM;
		*dst  += 1;

		u16 num = _Atoi(adrWord);
		**dst   = (num & 0x00ff);
		*dst   += 1;
		**dst   = (num & 0xff00) >> 8;
		*dst   += 1;
	}
	else
	{
		**dst  = COMET_TOKEN_ADR_LBL;
		*dst  += 1;

		p = adrWord;
		while((**dst = *p++) != '\0')
		{
			*dst += 1;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileStepNum(u8** src, u8** dst)
{
	u8  numWord[COMPILE_WORD_LENGTH];
	u8* p = numWord;

	do
	{
		*p++ = **src;
		*src += 1;

	} while( CompileIsChrNum(**src) == TRUE );

	*p++ = '\0';

	**dst   = COMET_TOKEN_NUM;
	*dst   += 1;

	u16 num = _Atoi(numWord);
	**dst   = (num & 0x00ff);
	*dst   += 1;
	**dst   = (num & 0xff00) >> 8;
	*dst   += 1;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsStrSpace(u8* src)
{
	while( CompileIsChrSpace(*src) == TRUE )
	{
		src++;
	}

	if( (*src == '\0') || (*src == ';') )
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsStrOp(u8* src)
{
	u8 opWord[COMPILE_WORD_LENGTH];

	if( CompileGetStr(src, opWord) == FALSE )
	{
		return FALSE;
	}

	//ߌe[uƔr܂
	u16 i;
	for(i=0; i<CASL_MAX_OP_KEYWORD_CNT; i++)
	{
		if( _Strcmp(opWord, CaslOpKeyWord[i].word) == 0 )
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsStrX(u8* src)
{
	u8 xWord[COMPILE_WORD_LENGTH];

	if( CompileGetStr(src, xWord) == FALSE )
	{
		return FALSE;
	}

	//Xe[uƔr܂
	u16 i;
	for(i=0; i<CASL_MAX_X_KEYWORD_CNT; i++)
	{
		if( _Strcmp(xWord, CaslxKeyWord[i].word) == 0 )
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsStrR(u8* src)
{
	u8 rWord[COMPILE_WORD_LENGTH];

	if( CompileGetStr(src, rWord) == FALSE )
	{
		return FALSE;
	}

	//GRe[uƔr܂
	u16 i;
	for(i=0; i<CASL_MAX_R_KEYWORD_CNT; i++)
	{
		if( _Strcmp(rWord, CaslrKeyWord[i].word) == 0 )
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//`FbN蔲Ȃ̂ŁAׂD揇ʂ͍Œɂ邱
EWRAM_CODE bool CompileIsStrAdr(u8* src)
{
	u8 adrWord[COMPILE_WORD_LENGTH];

	if( CompileGetStr(src, adrWord) == FALSE )
	{
		return FALSE;
	}

	//TODO Ȃ蔲
	if(adrWord[0] == '\0')
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsStrNum(u8* src)
{
	u16 len = _Length(src);
	u16 i;

	for(i=0; i<len; i++)
	{
		if( CompileIsChrNum(*src++) == FALSE )
		{
			return FALSE;
		}
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsChrName(u8 chr)
{
	return (chr >= 'A' && chr <= 'Z') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsChrNum(u8 chr)
{
	return (chr >= '0' && chr <= '9') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsChrSpace(u8 chr)
{
	return (chr == ' ') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsChrComma(u8 chr)
{
	return (chr == ',') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileIsChrNull(u8 chr)
{
	return (chr == '\0') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
//o܂
EWRAM_CODE bool CompileGetStr(u8* src, u8* dst)
{
	u16 cnt = 0;

	while( (CompileIsChrName(*src) == TRUE) || (CompileIsChrNum(*src) == TRUE) )
	{
		//̒`FbN
		if(cnt >= COMPILE_WORD_LENGTH)
		{
			return FALSE;
		}

		*dst++ = *src++;
		cnt++;
	}
	*dst++ = '\0';

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileInsertCode()
{
	ST_COMPILE* p = &Compile;
	ST_COMET*   c = &Comet;

	u8* src = p->tokenBuf;
	u8* dst = p->pMem;

	//L̍s͉I܂
	if(*src == 0x00)
	{
		return TRUE;
	}

	if(*src == COMET_TOKEN_LABEL)
	{
		src++;
		CompileRegLabelTable(src, ((u32)dst - (u32)&c->mem) >> 1);
		CompileStepNull(&src);
	}

	//op
	if(*src++ != COMET_TOKEN_OP)
	{
		return FALSE;
	}

	u8 opIdx  = *src++;
	u8 opCode = CaslOpKeyWord[opIdx].opCode;
	u8 token;

	switch( CaslOpKeyWord[opIdx].opType )
	{
	//START
	case OP_TYPE6:
		if(p->startFlag == TRUE)
		{
			//TODO START߂Qꍇ
			//TODO starẗʒúA擪H
			return FALSE;
		}
		p->startFlag = TRUE;
		break;

	//END
	case OP_TYPE7:
		p->endFlag = TRUE;
		break;

	//DS
	case OP_TYPE8:
		if(*src++ != COMET_TOKEN_NUM)
		{
			return FALSE;
		}
		u16 num  = *src++;

		num |= *src++ << 8;
		dst += num * 2;
		break;

	//DC
	case OP_TYPE9:
		if(*src++ != COMET_TOKEN_NUM)
		{
			return FALSE;
		}
		*dst++ = *src++;
		*dst++ = *src++;
		break;

	//IN,OUT
	case OP_TYPE10:
	case OP_TYPE11:
		//TODO
		break;

	//RPUSH,RPOP
	case OP_TYPE12:
	case OP_TYPE13:
		//TODO
		break;

	//none
	case OP_TYPE1:
		dst++;
		*dst++ = opCode;
		break;

	//r, adr ,xAr1, r2
	//r, adr ,x
	case OP_TYPE2:
	case OP_TYPE3:
		//param1
		if(*src++ != COMET_TOKEN_R)
		{
			return FALSE;
		}
		*dst++ = (*src++) << 4;
		*dst++ = opCode;

		//param2
		token = *src++;
		if(token == COMET_TOKEN_ADR_LBL)
		{
			CompileRefLabelTable(src, ((u32)dst - (u32)&c->mem) >> 1);
			CompileStepNull(&src);
			dst+=2;
		}
		else if(token == COMET_TOKEN_ADR_NUM)
		{
			*dst++ = *src++;
			*dst++ = *src++;
		}
		else if(token == COMET_TOKEN_R)
		{
			*(dst-2) |= *src++ << 4;
		}
		else
		{
			return FALSE;
		}

		//param3
		if(*src++ == COMET_TOKEN_X)
		{
			*(dst-2) |= *src++;
		}
		break;

	//r
	case OP_TYPE5:
		if(*src++ != COMET_TOKEN_R)
		{
			return FALSE;
		}
		*dst++ = (*src++) << 4;
		*dst++ = opCode;
		break;

	//adr ,x
	case OP_TYPE4:
		//param1
		token = *src++;
		if(token == COMET_TOKEN_ADR_LBL)
		{
			CompileRefLabelTable(src, ((u32)dst - (u32)&c->mem) >> 1);
			CompileStepNull(&src);
			dst+=2;
		}
		else if(token == COMET_TOKEN_ADR_NUM)
		{
			*dst++ = *src++;
			*dst++ = *src++;
		}
		else
		{
			return FALSE;
		}

		//param2
		if(*src++ == COMET_TOKEN_X)
		{
			*(dst-2) |= *src++;
		}
		break;

	default:
		DebugMessage("Err: CompileInsertCode param1");
		break;

	}// switch CaslOpKeyWord[opIdx].opType

	p->pMem = dst;
	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileRegLabelTable(u8* str, u16 adr)
{
	ST_COMPILE* p = &Compile;
	u16 cnt       = p->lblRegCnt;

	if(cnt >= COMPILE_MAX_REG_TABLE_CNT)
	{
		return FALSE;
	}

	_Strcpy(p->lblRegTbl[cnt].str, str);
	p->lblRegTbl[cnt].adr = adr;

	p->lblRegCnt++;

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileRefLabelTable(u8* str, u16 adr)
{
	ST_COMPILE* p = &Compile;
	u16 cnt       = p->lblRefCnt;

	if(cnt >= COMPILE_MAX_REG_TABLE_CNT)
	{
		return FALSE;
	}

	_Strcpy(p->lblRefTbl[cnt].str, str);
	p->lblRefTbl[cnt].adr = adr;

	p->lblRefCnt++;

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool CompileLinkLabelTable()
{
	ST_COMPILE* p = &Compile;
	u16 i, j;

	for(i=0; i<p->lblRefCnt; i++)
	{
		for(j=0; j<p->lblRegCnt; j++)
		{
			if( _Strcmp(p->lblRefTbl[i].str, p->lblRegTbl[j].str) == 0 )
			{
//				TRACEOUT("RefAdr:%04d  RegAdr:%04d\n", p->lblRefTbl[i].adr, p->lblRegTbl[i].adr);
				Comet.mem[ p->lblRefTbl[i].adr ] = p->lblRegTbl[j].adr;
				break;
			}
		}
	
		if(j >= p->lblRegCnt)
		{
			return FALSE;
		}
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileDebugToken(u16 y)
{
	ST_COMPILE* p = &Compile;

	TRACEOUT("--TokenDump---------------------------\n");
	TRACEOUT("[%02d]= ", y+1);

	u16 i;
	for(i=0; i<20; i++)
	{
		TRACEOUT("%02x ", p->tokenBuf[i]);
	}
	TRACEOUT("\n");
}
//---------------------------------------------------------------------------
EWRAM_CODE void CompileDebugMemory()
{
	ST_COMET* p = &Comet;

	TRACEOUT("--MemoryDump---------------------------\n");
	u16 i;
	for(i=0; i<10; i++)
	{
		TRACEOUT("[%04x] = %04x\n", i, p->mem[i]);
	}
	TRACEOUT("\n");
}




//---------------------------------------------------------------------------
EWRAM_CODE void CometInit()
{
	_Memset((u8*)&Comet, 0x00, sizeof(ST_COMET));

	ST_COMET* p = &Comet;

	p->PR = 0x0000;
	p->SP = COMET_START_SP_ADR;
	p->OF = 0;
	p->SF = 0;
	p->ZF = 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CometExec()
{
	ST_COMET* p = &Comet;
	u8  op, r1, r2;
	u16 adr;

	u32 ur   = 0;
	s32 sr   = 0;
	u16 flag = 0;

	TRACEOUT("\n\nComet Start! \n\n");

	for(;;)
	{
		CometDebugRegister();

		op  = CometGetOp();
		r1  = CometGetR1();
		r2  = CometGetR2();
		adr = CometGetAdr(r2);

		TRACEOUT("next op=[%02x]  r1=[%02x]  r2=[%02x]  adr=[%04x]\n\n", op, r1, r2, adr);

		switch(op)
		{
		//NOP
		case 0x00:
			p->PR += 1;
			break;

		//LD r,adr,x
		case 0x10:
			p->GR[r1] = ur = p->mem[adr];
			CometSetFRo1(ur);
			p->PR += 2;
			break;

		//ST r,adr,x
		case 0x11:
			p->mem[adr] = p->GR[r1];
			p->PR += 2;
			break;

		//LAD r,adr,x
		case 0x12:
			p->GR[r1] = adr;
			p->PR += 2;
			break;

		//LD r1,r2
		case 0x14:
			p->GR[r1] = ur = p->GR[r2];
			CometSetFRo1(ur);
			p->PR += 1;
			break;

		//ADDA r,adr,x
		case 0x20:
			p->GR[r1] = sr = p->GR[r1] + p->mem[adr];
			CometSetFRoS(sr);
			p->PR += 2;
			break;

		//SUBA r,adr,x
		case 0x21:
			p->GR[r1] = sr = p->GR[r1] - p->mem[adr];
			CometSetFRoS(sr);
			p->PR += 2;
			break;

		//ADDL r,adr,x
		case 0x22:
			p->GR[r1] = ur = p->GR[r1] + p->mem[adr];
			CometSetFRoL(ur);
			p->PR += 2;
			break;

		//SUBL r,adr,x
		case 0x23:
			p->GR[r1] = ur = p->GR[r1] - p->mem[adr];
			CometSetFRoL(ur);
			p->PR += 2;
			break;

		//ADDA r1,r2
		case 0x24:
			p->GR[r1] = sr = p->GR[r1] + p->GR[r2];
			CometSetFRoS(sr);
			p->PR += 1;
			break;

		//SUBA r1,r2
		case 0x25:
			p->GR[r1] = sr = p->GR[r1] - p->GR[r2];
			CometSetFRoS(sr);
			p->PR += 1;
			break;

		//ADDL r1,r2
		case 0x26:
			p->GR[r1] = ur = p->GR[r1] + p->GR[r2];
			CometSetFRoL(ur);
			p->PR += 1;
			break;

		//SUBL r1,r2
		case 0x27:
			p->GR[r1] = ur = p->GR[r1] - p->GR[r2];
			CometSetFRoL(ur);
			p->PR += 1;
			break;

		//AND r,adr,x
		case 0x30:
			p->GR[r1] = ur = p->GR[r1] & p->mem[adr];
			CometSetFRo1(ur);
			p->PR += 2;
			break;

		//OR r,adr,x
		case 0x31:
			p->GR[r1] = ur = p->GR[r1] | p->mem[adr];
			CometSetFRo1(ur);
			p->PR += 2;
			break;

		//XOR r,adr,x
		case 0x32:
			p->GR[r1] = ur = p->GR[r1] ^ p->mem[adr];
			CometSetFRo1(ur);
			p->PR += 2;
			break;

		//AND r1,r2
		case 0x34:
			p->GR[r1] = ur = p->GR[r1] & p->GR[r2];
			CometSetFRo1(ur);
			p->PR += 1;
			break;

		//OR r1,r2
		case 0x35:
			p->GR[r1] = ur = p->GR[r1] | p->GR[r2];
			CometSetFRo1(ur);
			p->PR += 1;
			break;

		//XOR r1,r2
		case 0x36:
			p->GR[r1] = ur = p->GR[r1] ^ p->GR[r2];
			CometSetFRo1(ur);
			p->PR += 1;
			break;

		//CPA r,adr,x
		case 0x40:
			sr = p->GR[r1] - p->mem[adr];
			CometSetFRo1(ur);
			p->PR += 2;
			break;

		//CPL r,adr,x
		case 0x41:
			ur = p->GR[r1] - p->mem[adr];
			CometSetFRo1(sr);
			p->PR += 2;
			break;

		//CPA r1,r2
		case 0x42:
			sr = p->GR[r1] - p->GR[r2];
			CometSetFRo1(ur);
			p->PR += 1;
			break;

		//CPL r1,r2
		case 0x43:
			ur = p->GR[r1] - p->GR[r2];
			CometSetFRo1(sr);
			p->PR += 1;
			break;

		//SLA r,adr,x
		case 0x50:
			flag = (p->GR[r1] << (p->mem[adr] - 1)) & 0x8000;
			p->GR[r1] = ur = (p->GR[r1] << p->mem[adr]) | (p->GR[r1] & 0x8000);
			CometSetFRo2(ur, flag);
			break;

		//SRA r,adr,x
		case 0x51:
			flag = (p->GR[r1] >> (p->mem[adr] - 1)) & 0x0001;
			p->GR[r1] = ur = (p->GR[r1] >> p->mem[adr]) | CometGetRevisionSRA(p->GR[r1], p->mem[adr]);
			CometSetFRo2(ur, flag);
			break;

		//SLL r,adr,x
		case 0x52:
			flag  = (p->GR[r1] << (p->mem[adr] - 1)) & 0x8000;
			p->GR[r1] = ur = (p->GR[r1] << p->mem[adr]);
			CometSetFRo2(ur, flag);
			break;

		//SRL r,adr,x
		case 0x53:
			flag = (p->GR[r1] >> (p->mem[adr] - 1)) & 0x0001;
			p->GR[r1] = ur = (p->GR[r1] >> p->mem[adr]);
			CometSetFRo2(ur, flag);
			break;

		//JMI adr,x
		case 0x61:
			p->PR = (p->SF == TRUE) ? p->mem[adr] : p->PR + 2;
			break;

		//JNZ adr,x
		case 0x62:
			p->PR = (p->ZF == FALSE) ? p->mem[adr] : p->PR + 2;
			break;

		//JZE adr,x
		case 0x63:
			p->PR = (p->ZF == TRUE) ? p->mem[adr] : p->PR + 2;
			break;

		//JUMP adr,x
		case 0x64:
			p->PR = p->mem[adr];
			break;

		//JPL adr,x
		case 0x65:
			p->PR = ((p->SF == FALSE) && (p->ZF == FALSE)) ? p->mem[adr] : p->PR + 2;
			break;
		
		//JOV adr,x
		case 0x66:
			p->PR = (p->OF == TRUE) ? p->mem[adr] : p->PR + 2;
			break;

		//PUSH adr,x
		case 0x70:
			p->mem[--p->SP] = adr;
			p->PR += 2;
			break;

		//POP r
		case 0x71:
			p->GR[r1] = p->mem[p->SP++];
			p->PR += 1;
			break;

		//CALL adr,x
		case 0x80:
			p->mem[--p->SP] = p->PR + 2;
			p->PR = p->mem[adr];
			break;

		//RET
			case 0x81:
			if(p->SP == COMET_START_SP_ADR)
			{
				goto End;
			}
			p->PR = p->mem[p->SP++];
			break;

		default:
			DebugMessage("Err: CometExec Op");
			break;
		}
	}

End:

CompileDebugMemory();

for(;;){}
}
//---------------------------------------------------------------------------
EWRAM_CODE void CometSetFRo1(u16 r)
{
	ST_COMET* p = &Comet;

	p->OF = 0;
	p->ZF = (r == 0x0000) ? 1 : 0;
	p->SF = (r &  0x8000) ? 1 : 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CometSetFRo2(u16 r, u16 of)
{
	ST_COMET* p = &Comet;

	p->OF = (of != 0) ? 1 : 0;
	p->ZF = (r == 0x0000) ? 1 : 0;
	p->SF = (r &  0x8000) ? 1 : 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CometSetFRoS(s32 r)
{
	ST_COMET* p = &Comet;

	p->OF = (r < -32768 || 32767 < r) ? 1 : 0;
	p->ZF = (r == 0x0000) ? 1 : 0;
	p->SF = (r &  0x8000) ? 1 : 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CometSetFRoL(u32 r)
{
	ST_COMET* p = &Comet;

	p->OF = (r >   65535) ? 1 : 0;
	p->ZF = (r == 0x0000) ? 1 : 0;
	p->SF = (r &  0x8000) ? 1 : 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 CometGetOp()
{
	ST_COMET* p = &Comet;
	return p->mem[ p->PR ] >> 8;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 CometGetR1()
{
	ST_COMET* p = &Comet;
	return ((p->mem[ p->PR ]) >> 4) & 0x000f;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 CometGetR2()
{
	ST_COMET* p = &Comet;
	return p->mem[ p->PR ] & 0x000f;
}
//---------------------------------------------------------------------------
EWRAM_CODE u16 CometGetAdr(u8 x)
{
	ST_COMET* p = &Comet;
	u16 tmp     = (x == 0) ? 0 : p->GR[x];

	return p->mem[ p->PR+1 + tmp ];
}
//---------------------------------------------------------------------------
//SRA߂̕␳֐orz
EWRAM_CODE u16 CometGetRevisionSRA(u16 r, u16 cnt)
{
	u16 i;
	u16 mask = r & 0x8000;

	for(i=0; i<cnt; i++)
	{
		r >>= 1;
		r  |= mask;
	}

	return r;
}
//---------------------------------------------------------------------------
EWRAM_CODE void CometDebugRegister()
{
	ST_COMET* p = &Comet;

	TRACEOUT("GR%1d=%04x ", 0, p->GR[0]);
	TRACEOUT("GR%1d=%04x ", 1, p->GR[1]);
	TRACEOUT("GR%1d=%04x ", 2, p->GR[2]);
	TRACEOUT("GR%1d=%04x ", 3, p->GR[3]);
	TRACEOUT("\n");

	TRACEOUT("GR%1d=%04x ", 4, p->GR[4]);
	TRACEOUT("GR%1d=%04x ", 5, p->GR[5]);
	TRACEOUT("GR%1d=%04x ", 6, p->GR[6]);
	TRACEOUT("GR%1d=%04x ", 7, p->GR[7]);
	TRACEOUT("\n");

	TRACEOUT("SP=%04x ", p->SP);
	TRACEOUT("PR=%04x ", p->PR);
	TRACEOUT("(OF=%1x SF=%1x ZF=%1x)\n", p->OF, p->SF, p->ZF);
}
